/***
 * Copyright (C) 1996-98. by Miroslav Zagorac <zaga@tirnanog.cc.fer.hr>
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "../data/objects_def.h"
#include "../lib/lib_debug.h"

#include "cave.h"
#include "cave_data.h"
#include "graphics.h"
#include "level_editor.h"
#include "main.h"
#include "objects.h"
#include "player.h"


static struct COORDINATE counter_scroll;

struct COORDINATE cave_max, cave_offset, cave_out;
struct CAVE_DATA  cave;


#ifdef ANSI_FUNC

void coords_clear (struct COORDINATE *coord)
#else

void coords_clear (coord)
   struct COORDINATE *coord;
#endif
{
   coord->x = coord->y = 0;
}


#ifdef ANSI_FUNC

void cave_set_position (short position, unchar objects, unchar attr, unchar anim)
#else
   
void cave_set_position (position, objects, attr, anim)
   short position;
   unchar objects;
   unchar attr;
   unchar anim;
#endif
{
   cave.data[position] = objects;
   cave.attr[position] = attr;
   cave.anim[position] = anim;
}


#ifdef ANSI_FUNC

void cave_exchange_position (short new_position, short old_position)
#else
   
void cave_exchange_position (new_position, old_position)
   short new_position;
   short old_position;
#endif
{
   cave_set_position (new_position, cave.data[old_position], cave.attr[old_position], cave.anim[old_position]);
   cave_set_position (old_position, SPACE, 0, 0);
}


#ifdef ANSI_FUNC

void cave_next (void)
#else
   
void cave_next ()
#endif
{
   debug (("{%s()", __FUNCTION__))

   /***
    * Povratak sa Intermission labirinta.
    */
   if (player_data[game_data.current_player].room >= CAVE_MAXROOM)
      player_data[game_data.current_player].room = (player_data[game_data.current_player].room - CAVE_MAXROOM + 1) * 4;
   /***
    * Da li je vrijeme za Intermission labirint?
    */
   else if (! ((player_data[game_data.current_player].room + 1) % 4))
   {
      player_data[game_data.current_player].room = CAVE_MAXROOM + player_data[game_data.current_player].room / 4;

      debug_noreturn
   }

   if (++player_data[game_data.current_player].room >= CAVE_MAXROOM)
   {
      player_data[game_data.current_player].room = 0;

      /***
       * Nivoi se povecavaju sve do zadnjega...
       */
      if (player_data[game_data.current_player].level < CAVE_MAXLEVEL)
         player_data[game_data.current_player].level++;
   }

   debug_endfunc
}


#ifdef ANSI_FUNC

short cave_load (short room, short level)
#else

short cave_load (room, level)
   short room;
   short level;
#endif
{
   short retval = FUNC_RET_OK, i;

   debug (("{%s(%d, %d)", __FUNCTION__, room, level))

   srandom (time (NULL));

   coords_clear (&cave_max);
   coords_clear (&counter_scroll);

   (void) memset (cave.attr, 0, CAVE_SIZE);
   (void) memset (cave.anim, 0, CAVE_SIZE);

   if ((retval = level_load (room, level, &cave)) != FUNC_RET_OK)
      debug_return ("} == %d", retval)

   for (i = 0; i < CAVE_SIZE; i++)
   {
      short x = i % CAVE_WIDTH, y = i / CAVE_WIDTH;

      /***
       * Gdje je pocetna pozicija igraca?
       */
      if (cave.data[i] == PRE_ROCKFORD_1)
      {
         rockford_coord.x = x;
         rockford_coord.y = y;
      }

      /***
       * Gdje je izlaz?
       */
      if (cave.data[i] == PRE_OUTBOX)
      {
         cave_out.x = x;
         cave_out.y = y;
      }

      /***
       * Ovdje se trazi maksimalna velicina labirinta, koja ne mora biti jednaka
       * maksimalno mogucoj velicini istog.
       */
      if (cave.data[i] != SPACE)
      {
         cave_max.x = (cave_max.x < x) ? x : cave_max.x;
         cave_max.y = (cave_max.y < y) ? y : cave_max.y;
      }
   }

   cave_max.x++;
   cave_max.y++;

   game_data.game_status     &= GAME_DEMO_MODE;
   game_data.diamonds         = 0;
   game_data.scroll_step      = game_data.scroll_basic_step;
   game_data.max_amoeba_size  = cave_max.x * cave_max.y * 200.0 / 880.0;

   /***
    * Na pocetku igre, treba se pozicionirati negdje u sredisnji dio ekrana.
    */
   game_data.border_dist_x = CAVE_VWIDTH / 2 - 1;
   game_data.border_dist_y = CAVE_VHEIGHT / 2 - 1;

   debug_return ("} == %d", retval)
}


#ifdef ANSI_FUNC

void cave_mask (short mask_flag)
#else
   
void cave_mask (short mask_flag)
   short mask_flag;
#endif
{
   static unchar mask_data[CAVE_SIZE], counter = 0;
   static short  status = 0;
   short         mask_number = CAVE_SIZE, x, y;

   debug (("{%s(0x%04x)", __FUNCTION__, mask_flag))

   if (mask_flag & (CAVE_MASK_DEC | CAVE_MASK_INC))
   {
      (void) memset (mask_data, (mask_flag & CAVE_MASK_DEC) ? 0xf0 : 0x00, CAVE_SIZE);

      status = mask_flag;

      debug_noreturn
   }
   else if (mask_flag & CAVE_MASK_CLEAR)
   {
      (void) memset (mask_data, (status & CAVE_MASK_DEC) ? 0x00 : 0xf0, CAVE_SIZE);

      status = mask_flag;
   }

   if (! (status & CAVE_MASK_CLEAR))
      for (x = mask_number = 0, counter++; x < CAVE_SIZE; x++)
      {
         if (counter & 0x01)
         {
            if ((random () % 100) < 5)
	    {
               if (status & CAVE_MASK_DEC)
                  mask_data[x] &= (random () & 0xf0) | 0x0f;
               else
                  mask_data[x] |= random () & 0xf0;
	    }

            mask_data[x] = (mask_data[x] & 0xf0) | ((mask_data[x] + 1) & 0x07);
         }

         mask_number += (mask_data[x] & 0xf0) ? 1 : 0;
      }

   if (! mask_number)
      debug_noreturn

   /***
    * mask_data[]:
    *   bits 7-4  -- po jedan bit za svaku poziciju velicine 16x16 unutar podrucja 32x32
    *   bits 3-0  -- brojac za broj slike (animacije)
    */
   for (x = 0; x < (CAVE_VWIDTH + (counter_scroll.x ? 1 : 0)); x++)
      for (y = 0; y < (CAVE_VHEIGHT + (counter_scroll.y ? 1 : 0)); y++)
      {
         short object = mask_data[CAVE_POSITION (cave_offset.x + x, cave_offset.y + y)];
         short x_pos  = x * 32 + ((counter_scroll.x > 0) ? -32 : 0) + counter_scroll.x;
         short y_pos  = (y + 1) * 32 + ((counter_scroll.y > 0) ? -32 : 0) + counter_scroll.y;

         if (object & 0x10)
            graph_putbox (x_pos, y_pos, 16, 16, get_object_image (WALL_PART, -(object & 0x07)));

         if (object & 0x20)
            graph_putbox (x_pos + 16, y_pos, 16, 16, get_object_image (WALL_PART, -(object & 0x07)));

         if (object & 0x40)
            graph_putbox (x_pos, y_pos + 16, 16, 16, get_object_image (WALL_PART, -(object & 0x07)));

         if (object & 0x80)
            graph_putbox (x_pos + 16, y_pos + 16, 16, 16, get_object_image (WALL_PART, -(object & 0x07)));
      }

   debug_endfunc
}


#ifdef ANSI_FUNC

void cave_show (void)
#else

void cave_show ()
#endif
{
   static struct COORDINATE cave_scroll_direction = { 0, 0 };
   short                    x, y;

   debug (("{%s()", __FUNCTION__))

   /***
    * Pomicanje u horizontalnom smijeru.
    */
   if (! counter_scroll.x)
   {
      /***
       * Treba li pomaknuti sliku labirinta?
       */
      if ((cave_offset.x + CAVE_SCROLL_DISTANCE) > rockford_coord.x)
         cave_scroll_direction.x = -1;
      else if ((cave_offset.x + (CAVE_VWIDTH - CAVE_SCROLL_DISTANCE)) <= rockford_coord.x)
         cave_scroll_direction.x = 1;

      /***
       * Ne smije se dozvoliti izlazenje izvan granica labirinta...
       */
      if ((!cave_offset.x && (cave_scroll_direction.x < 0)) || ((cave_offset.x >= (cave_max.x - CAVE_VWIDTH)) && (cave_scroll_direction.x > 0)))
         cave_scroll_direction.x = 0;

      /***
       * Pomicanje se zaustavlja tek na odredjenom razmaku kraja ekrana od Dasha.
       */
      if (cave_scroll_direction.x)
         if (((cave_offset.x + game_data.border_dist_x) <= rockford_coord.x) && ((cave_offset.x + (CAVE_VWIDTH - game_data.border_dist_x)) > rockford_coord.x))
            cave_scroll_direction.x = 0;

      if (cave_scroll_direction.x > 0)
         counter_scroll.x = 32;
      else if (cave_scroll_direction.x < 0)
      {
         counter_scroll.x = -32;
         cave_offset.x--;
      }
   }

   if (counter_scroll.x)
   {
      if (counter_scroll.x > 0)
      {
         counter_scroll.x -= (counter_scroll.x < game_data.scroll_step) ? counter_scroll.x : game_data.scroll_step;

         if (! counter_scroll.x)
            cave_offset.x++;
      }
      else
         counter_scroll.x += (counter_scroll.x > -game_data.scroll_step) ? -counter_scroll.x : game_data.scroll_step;
   }

   /***
    * Pomicanje u vertikalnom smijeru.
    */
   if (! counter_scroll.y)
   {
      /***
       * Treba li pomaknuti sliku labirinta?
       */
      if ((cave_offset.y + CAVE_SCROLL_DISTANCE) > rockford_coord.y)
         cave_scroll_direction.y = -1;
      else if ((cave_offset.y + (CAVE_VHEIGHT - CAVE_SCROLL_DISTANCE)) <= rockford_coord.y)
         cave_scroll_direction.y = 1;

      /***
       * Ne smije se dozvoliti izlazenje izvan granica labirinta...
       */
      if ((!cave_offset.y && (cave_scroll_direction.y < 0)) || ((cave_offset.y >= (cave_max.y - CAVE_VHEIGHT)) && (cave_scroll_direction.y > 0)))
         cave_scroll_direction.y = 0;

      /***
       * Pomicanje se zaustavlja tek na odredjenom razmaku kraja ekrana od Dasha.
       */
      if (cave_scroll_direction.y)
         if (((cave_offset.y + game_data.border_dist_y) <= rockford_coord.y) && ((cave_offset.y + (CAVE_VHEIGHT - game_data.border_dist_y)) > rockford_coord.y))
            cave_scroll_direction.y = 0;

      if (cave_scroll_direction.y > 0)
         counter_scroll.y = 32;
      else if (cave_scroll_direction.y < 0)
      {
         counter_scroll.y = -32;
         cave_offset.y--;
      }
   }

   if (counter_scroll.y)
   {
      if (counter_scroll.y > 0)
      {
         counter_scroll.y -= (counter_scroll.y < game_data.scroll_step) ? counter_scroll.y : game_data.scroll_step;

         if (! counter_scroll.y)
            cave_offset.y++;
      }
      else
         counter_scroll.y += (counter_scroll.y > -game_data.scroll_step) ? -counter_scroll.y : game_data.scroll_step;
   }

   /***
    * Ovdje se prikazuje labirint na ekranu.
    */
   graph_setclippingwindow (0, 32, WINDOW_WIDTH - 1, WINDOW_HEIGHT - 1);

   for (x = 0; x < (CAVE_VWIDTH + (counter_scroll.x ? 1 : 0)); x++)
      for (y = 0; y < (CAVE_VHEIGHT + (counter_scroll.y ? 1 : 0)); y++)
      {
         short position = CAVE_POSITION (cave_offset.x + x, cave_offset.y + y);

         graph_putbox (x * 32 + ((counter_scroll.x > 0) ? -32 : 0) + counter_scroll.x, (y + 1) * 32 + ((counter_scroll.y > 0) ? -32 : 0) + counter_scroll.y, 32, 32, get_object_image (cave.data[position], position));
      }

   cave_mask (0);
   graph_setclippingwindow (0, 0, WINDOW_WIDTH - 1, WINDOW_HEIGHT - 1);

   debug_endfunc
}


#ifdef ANSI_FUNC

void cave_mask_set (short mask_flag, time_t move_time)
#else

void cave_mask_set (mask_flag, move_time)
   short mask_flag;
   time_t move_time;
#endif
{
   short  init_flag = 0;
   time_t start_time = time (NULL);

   debug (("{%s(0x%04x, %ld)", __FUNCTION__, mask_flag, move_time))

   cave_mask (mask_flag);
   game_timer (1);

   while ((time (NULL) - start_time) < move_time)
   {
      game_timer (0);

      if (game_data.game_status & GAME_OBJECT_ANIMATE)
         objects_animate ();

      cave_show ();
      graph_copyscreen ();

      if (!init_flag && (mask_flag & CAVE_MASK_FADE))
         graph_palette_fade (init_flag = 4);
   }

   cave_mask (CAVE_MASK_CLEAR);
   cave_show ();
   graph_copyscreen ();

   debug_endfunc
}
