#ifndef _CAVE_DATA_H_
#define _CAVE_DATA_H_


#define CAVE_RAW_MAXSIZE   256

enum LINE_directions
{
   LINE_DIR_UP,
   LINE_DIR_UPRIGHT,
   LINE_DIR_RIGHT,
   LINE_DIR_DOWNRIGHT,
   LINE_DIR_DOWN,
   LINE_DIR_DOWNLEFT,
   LINE_DIR_LEFT,
   LINE_DIR_UPLEFT
};

enum OBJECT_definitions
{
   SPACE,		/* 0x00 Space                                                     */
   DIRT,		/* 0x01 Dirt                                                      */
   BRICK_WALL,		/* 0x02 Brick wall                                                */
   MAGIC_WALL,		/* 0x03 Magic wall                                                */
   PRE_OUTBOX,		/* 0x04 Pre outbox (ie before it starts flashing)                 */
   FLASH_OUTBOX,	/* 0x05 Flashing outbox                                           */
   _1,			/* 0x06                                                           */
   WALL,		/* 0x07 Steel wall                                                */
   FIREFLY_1,		/* 0x08 Firefly position 1                                        */
   FIREFLY_2,		/* 0x09 Firefly position 2                                        */
   FIREFLY_3,		/* 0x0a Firefly position 3                                        */
   FIREFLY_4,		/* 0x0b Firefly position 4                                        */
   FIREFLY_1_S,		/* 0x0c Firefly position 1, scanned this frame                    */
   FIREFLY_2_S,		/* 0x0d Firefly position 2, scanned this frame                    */
   FIREFLY_3_S,		/* 0x0e Firefly position 3, scanned this frame                    */
   FIREFLY_4_S,		/* 0x0f Firefly position 4, scanned this frame                    */
   BOULDER,		/* 0x10 Boulder, stationary                                       */
   BOULDER_S,		/* 0x11 Boulder, stationary, scanned this frame                   */
   BOULDER_FALL,	/* 0x12 Boulder, falling                                          */
   BOULDER_FALL_S,	/* 0x13 Boulder, falling, scanned this frame                      */
   DIAMOND,		/* 0x14 Diamond, stationary                                       */
   DIAMOND_S,		/* 0x15 Diamond, stationary, scanned this frame                   */
   DIAMOND_FALL,	/* 0x16 Diamond, falling                                          */
   DIAMOND_FALL_S,	/* 0x17 Diamond, falling, scanned this frame                      */
   _2,			/* 0x18                                                           */
   _3,			/* 0x19                                                           */
   _4,			/* 0x1a                                                           */
   EXPLODE_S0,		/* 0x1b Explode to Space stage 0                                  */
   EXPLODE_S1,		/* 0x1c Explode to Space stage 1                                  */
   EXPLODE_S2,		/* 0x1d Explode to Space stage 2                                  */
   EXPLODE_S3,		/* 0x1e Explode to Space stage 3                                  */
   EXPLODE_S4,		/* 0x1f Explode to Space stage 4                                  */
   EXPLODE_D0,		/* 0x20 Explode to Diamond stage 0                                */
   EXPLODE_D1,		/* 0x21 Explode to Diamond stage 1                                */
   EXPLODE_D2,		/* 0x22 Explode to Diamond stage 2                                */
   EXPLODE_D3,		/* 0x23 Explode to Diamond stage 3                                */
   EXPLODE_D4,		/* 0x24 Explode to Diamond stage 4                                */
   PRE_ROCKFORD_1,	/* 0x25 Pre Rockford stage 1 (ie when inbox morphs into Rockford) */
   PRE_ROCKFORD_2,	/* 0x26 Pre Rockford stage 2                                      */
   PRE_ROCKFORD_3,	/* 0x27 Pre Rockford stage 3                                      */
   PRE_ROCKFORD_4,	/* 0x28 Pre Rockford stage 4                                      */
   _5,			/* 0x29                                                           */
   _6,			/* 0x2a                                                           */
   _7,			/* 0x2b                                                           */
   _8,			/* 0x2c                                                           */
   _9,			/* 0x2d                                                           */
   _10,			/* 0x2e                                                           */
   _11,			/* 0x2f                                                           */
   BUTTERFLY_1,		/* 0x30 Butterfly position 1                                      */
   BUTTERFLY_2,		/* 0x31 Butterfly position 2                                      */
   BUTTERFLY_3,		/* 0x32 Butterfly position 3                                      */
   BUTTERFLY_4,		/* 0x33 Butterfly position 4                                      */
   BUTTERFLY_1_S,	/* 0x34 Butterfly position 1, scanned this frame                  */
   BUTTERFLY_2_S,	/* 0x35 Butterfly position 2, scanned this frame                  */
   BUTTERFLY_3_S,	/* 0x36 Butterfly position 3, scanned this frame                  */
   BUTTERFLY_4_S,	/* 0x37 Butterfly position 4, scanned this frame                  */
   ROCKFORD,		/* 0x38 Rockford                                                  */
   ROCKFORD_S,		/* 0x39 Rockford, scanned this frame                              */
   AMOEBA,		/* 0x3a Amoeba                                                    */
   AMOEBA_S,		/* 0x3b Amoeba, scanned this frame                                */
   _12,			/* 0x3c                                                           */
   _13,			/* 0x3d                                                           */
   _14,			/* 0x3e                                                           */
   _15,			/* 0x3f                                                           */

   ROCKFORD_RANDOM,
   ROCKFORD_0,
   ROCKFORD_1,
   ROCKFORD_2,
   ROCKFORD_3,
   ROCKFORD_R,
   ROCKFORD_L,
   WALL_PART,
   PRE_MAGIC_WALL,
   POST_MAGIC_WALL,
   UNKNOWN
};

struct CAVE_RAW_DATA
{
   unchar room;			   /* 0x00: Cave number                                                */
   unchar wall_time;		   /* 0x01: Magic wall milling time, also max amoeba time at 3% growth */
   unchar dvalue_init;		   /* 0x02: Initial diamond value                                      */
   unchar dvalue_extra;		   /* 0x03: Extra diamond value                                        */
   unchar random[5];		   /* 0x04: Initial randomiser seed value for difficulty levels 1-5    */
   unchar diamonds[5];		   /* 0x09: Diamonds needed on difficulty levels 1-5                   */
   unchar time[5];		   /* 0x0e: Cave time for difficulty levels 1-5                        */
   unchar bg1_color;		   /* 0x13: Background colour 1                                        */
   unchar bg2_color;		   /* 0x14: Background colour 2                                        */
   unchar fg_color;		   /* 0x15: Foreground colour                                          */
   unchar unused[2];		   /* 0x16: unused                                                     */
   unchar object_random[4];	   /* 0x18: Random object numbers 1-4                                  */
   unchar object_probability[4];   /* 0x1c: Probability of objects 1-4                                 */
   unchar data[CAVE_SIZE];	   /* 0x20 onwards: cave data, deliminated with 0xFF                   */
};


#if defined (__STDC__) || defined (__cplusplus) || defined (NeedFunctionPrototypes)
#  define __PARMS(s) s
#else
#  define __PARMS(s) ()
#endif

void  store_single_object   __PARMS((unchar object, unchar column, unchar row));
void  draw_line             __PARMS((unchar object, unchar column, unchar row, unchar length, unchar direction));
void  draw_filled_rectangle __PARMS((unchar object, unchar column, unchar row, unchar length, unchar height, unchar object_fill));
void  draw_rectangle        __PARMS((unchar object, unchar column, unchar row, unchar length, unchar height));
void  next_random           __PARMS((unchar *random_seed_1, unchar *random_seed_2));
short cave_decode           __PARMS((short room, short level, struct CAVE_DATA *cave_data));

#ifdef DEBUG
void  caves_all_check       __PARMS((void));
#endif

#undef __PARMS
#endif /* _CAVE_DATA_H_ */
